﻿library versions;

{$DEFINE NOFORMS}

uses
  Classes,
  SysUtils,
  Windows,
  ShellApi,
  System.Net.HttpClient,
  System.Generics.Collections,
  System.Generics.Defaults,
  ObjectIni in 'ObjectIni.pas',
  ObjectLines in 'ObjectLines.pas',
  UnitStrings in 'UnitStrings.pas';

  {$E wfx}

{$IFDEF WIN64}
  {$E w64}
  {$IFNDEF VER230}
    {$E wfx64}
  {$ENDIF}
{$ENDIF}

{$R Icons.res}
{$R versions.res}

var
  HTTPClient: THTTPClient;
  FileDate : TFileTime;
  PathExe : string;
  FileList : TStringList;
  CurrentPath : string;
  FLIndex : Integer;
  CachePages: TDictionary<string, string>;
  UseCache: Boolean = True;
  UseLowerCase: Boolean = True;

function GetPage(const url : string) : string;
var
  R: IHTTPResponse;
begin
  try
    R := HTTPClient.Get(url);
    if R.StatusCode = 200 then
      Result := R.ContentAsString(TEncoding.ANSI);
  except
    on E: Exception do
    begin
      Result := '';
      //TCShowMessage('Get Page Error', url + #13 + E.ClassName + #13 + E.Message);
    end;
  end;
end;
//______________________________________________________________________________
function DateTimeToFileTime(MyDateTime : TDateTime): TFileTime;
var
  MyFileAge : Integer;
  MyLocalFileTime : _FILETIME;
begin
  MyFileAge := DateTimeToFileDate(MyDateTime);
  DosDateTimeToFileTime(LongRec(MyFileAge).Hi, LongRec(MyFileAge).Lo, MyLocalFileTime);
  LocalFileTimeToFileTime(MyLocalFileTime, Result);
end;
//______________________________________________________________________________
function FsInit(PluginNr : Integer; pProgressProc : TProgressProc; pLogProc : TLogProc; pRequestProc : TRequestProc) : Integer; stdcall;
begin
  Result := -1;
end;

function FsInitW(PluginNr: Integer; pProgressProcW: tProgressProcW;
  pLogProcW: tLogProcW; pRequestProcW: tRequestProcW): Integer; stdcall;
begin
  PluginNumber := PluginNr;
  EnterText := pRequestProcW;
  ProgressBar := pProgressProcW;
  LogProc := pLogProcW;
  Result := 0;
  FileDate := DateTimeToFileTime(Now);
end;
//______________________________________________________________________________
procedure BuildFindData(var FD : TWin32FindDataW; FileName : string; Size: Integer);
begin
  FillChar(FD, SizeOf(FD), 0);
  StrCopy(FD.cFileName, PChar(FileName));
  FD.nFileSizeLow := Size;
  FD.ftCreationTime := FileDate;
  FD.ftLastWriteTime := FileDate;
  FD.ftLastAccessTime := FileDate;
end;
//______________________________________________________________________________
procedure FindFileList(var sl : TStringList; Mask : string);
var
  sr : TSearchRec;
begin
  if FindFirst(Mask, faAnyFile, sr) = 0 then begin
    sl.AddObject(sr.Name, TObject(Int64Rec(sr.Size).Lo));
    while FindNext(sr) = 0 do sl.AddObject(sr.Name, TObject(Int64Rec(sr.Size).Lo));
  end;
end;
//______________________________________________________________________________
procedure RootList(var sl : TStringList);
begin
  FindFileList(sl, PathExe + '*.list');
  sl.Add(Strings[1]); //Choose Language
  sl.Add(Strings[10]); //Edit List
end;
//______________________________________________________________________________
procedure LangList(var sl : TStringList);
var
  i : Integer;
begin
  FindFileList(sl, PathExe + 'ver_*.lng');
  for i := 0 to sl.Count - 1 do sl[i] := ChangeFileExt(sl[i], '');
end;
//______________________________________________________________________________
procedure PageList(var sl : TStringList);
var
  s : string;
begin
  s := CurrentPath;
  Delete(s, 1, 1);
  sl.LoadFromFile(PathExe + s);
end;
//______________________________________________________________________________
procedure CreateFileList;
begin
  FileList.Clear;
  if CurrentPath = '\' then RootList(FileList)
  else if Pos('.list', CurrentPath) > 0 then PageList(FileList)
  else if CurrentPath = '\' + Strings[1] then LangList(FileList);
end;
//______________________________________________________________________________
function InputParams(title, line, params : string) : string;
var
  i : Integer;
  sl : TStringList;
  s : string;
  Text : string;
  C : Integer;
begin
  Result := '';
  s := '';

  sl := TStringList.Create;
  try
    Split(sl, Params, ',');
    C := sl.Count;
    for i := 0 to C - 1 do begin
      s := '';
      Text := Strings[15] + #13#10 + sl[i];
      if C > 1 then Text := Text + Format(' (%d/%d)', [i + 1, C]);
      if not Input(title, Text, s) then Exit;
      Line := Replace(Line, Format('[%.2d]', [i + 1]), s);
    end;
  finally
    FreeAndNil(sl);
  end;

  Result := line;

//  if not Input(title, Strings[15] + #13#10 + '[' + params + ']', s) then Exit;
//  Result := line;
//  sl := TStringList.Create;
//  try
//    Split(sl, s, ',');
//    for i := 1 to sl.Count do Result := Replace(Result, Format('[%.2d]', [i]), sl[i - 1]);
//  finally
//    FreeAndNil(sl);
//  end;
end;
//______________________________________________________________________________
procedure LoadPage(ofn : string);
var
  s : string;
  i : Integer;
  sl : TStringList;
  page : string;
  line : TStringList;
begin
  //si il s'agit d'un page d'erreur, on sort
  if ofn = Strings[13] then Exit;

  sl := TStringList.Create;
  try
    //recherche du non de fichier .list associé à la recherche
    page := CurrentPath;
    Delete(page, 1, 1);
    sl.LoadFromFile(PathExe + page);

    //suppression de l'extension de ofn
    ofn := ChangeFileExt(ofn, '');

    //On vire le numéro de version qui suit le nom
    i := Pos('(', ofn);
    if i > 0 then Delete(ofn, i, Length(ofn));
    ofn := Trim(ofn);

    //recherche de la ligne du fichier .list contenant le nom de fichier sélectionné
    for i := 0 to sl.Count - 1 do begin
      if Pos(ofn + '|', sl[i]) = 1 then begin
        //On a trouvé la bonne ligne
        line := TStringList.Create;
        try
          Split(line, sl[i], '|');
          if line.Count = 3 then s := InputParams(line[0], line[1], line[2])
          else s := line[1];
        finally
          FreeAndNil(line);
        end;
        if s <> '' then ShellExecute(0, PChar('open'), PChar(s), PChar(''), PChar(''), SW_SHOWNORMAL);
        Exit;
      end;
    end;
  finally
    sl.Free;
  end;
end;
//______________________________________________________________________________
procedure AddSearch;
var
  i : Integer;
  sl : TStringList;
  s : string;
  Line : string;
  FileName : string;
begin

  FileName := '';
  s := '';
  if not Input(Strings[10], Strings[0], s) then Exit;
  if ExtractFileExt(s) <> '.list' then s := s + '.list';

  s := Trim(s);
  i := 1;
  while i < Length(s) do
    if s[i] in ['0'..'9', 'a'..'z', 'A'..'Z', '_', ' ', '.'] then Inc(i)
    else Delete(s, i, 1);
  if s = '' then Exit;

  FileName := s;

  s := '';
  if not Input(Strings[10], Strings[4], s) then Exit;
  Line := Replace(s, '(', '');
  Line := Replace(Line, ')', '');

  s := '';
  if not Input(Strings[10], Strings[5], s) then Exit;
  Line := Line + '|' + s;

  s := '';
  if not Input(Strings[10], Strings[6], s) then Exit;
  Line := Line + '|' + s;

    s := '';
  if Input(Strings[10], Strings[7], s) then begin
    Line := Line + '|' + s;

    s := '';
    if not Input(Strings[10], Strings[8], s) then Exit;
    Line := Line + '|' + s;

    s := '';
    if not Input(Strings[10], Strings[9], s) then Exit;
    Line := Line + '|' + s;
  end;

  sl := TStringList.Create;
  try
    if FileExists(PathExe + FileName) then
      sl.LoadFromFile(PathExe + FileName)
    else
      sl.DefaultEncoding := TEncoding.Unicode;
    sl.Add(Line);
    {
    // Convert file to Unicode if not
    if Assigned(sl.Encoding) and ((sl.Encoding.CodePage = 1200) or
      (sl.Encoding.CodePage = 1201) or (sl.Encoding.CodePage = 65001)) then
      sl.SaveToFile(PathExe + FileName)
    else
      sl.SaveToFile(PathExe + FileName, TEncoding.Unicode);
    }
    sl.SaveToFile(PathExe + FileName)
  finally
    FreeAndNil(sl);
  end;

end;
//______________________________________________________________________________
function FsExecuteFile(MainWin : THandle; RemoteName, Verb : PAnsiChar) : Integer; stdcall;
begin
  Result := FS_EXEC_ERROR
end;

function FsExecuteFileW(MainWin:THandle; RemoteName, Verb:PWideChar):integer; stdcall;
var
  Name : string;
  ofn : string;
  vb : string;
  ext : string;
begin

  Name := StrPas(RemoteName);
  Result := FS_EXEC_SYMLINK;

  ofn := StrPas(RemoteName);
  ofn := DeleteUntilLast(ofn, '\');

  vb := strpas(Verb);
  if Pos('quote', vb) = 1 then begin
    Exit;
  end;

  ext := ExtractFileExt(Name);

  Delete(Name, 1, 1);

  if ext = '.list' then begin
    if vb = 'properties' then begin
      ///edit list
      ShellExecute(0, PChar('open'), PChar('notepad'), PChar(PathExe + ofn), PChar(PathExe), SW_SHOWNORMAL);
      Result := FS_EXEC_OK;
    end
    else begin
      //Connect and download the selected .list
      FileList.Clear;
      StrPCopy(RemoteName, '\' + Name);
      CurrentPath := StrPas(RemoteName);
    end;
  end
  else if Pos('.list\', Name) > 0 then begin
    //Chargement de l'url dans le navigateur
    Result := FS_EXEC_OK;
    LoadPage(Ofn);
  end
  else if ofn = Strings[1] then StrPCopy(RemoteName, '\' + Strings[1]) //Language
  else if RemoteName = '\' + Strings[1] + '\' + ofn then begin //language 2
    Ini.SetValue('Language', ofn);
    Strings.SetLanguage(PathExe + ofn + '.lng');
    StrPCopy(RemoteName, '\');
  end
  else if Name = Strings[10] then begin
    Result := FS_EXEC_OK;
    AddSearch;
  end
  else Result := FS_EXEC_YOURSELF;

end;
//______________________________________________________________________________
function DoLowerCase(const Str: string; UseLocale: Boolean = False): string; inline;
begin
  if not UseLowerCase then
    Exit(Str);

  if UseLocale then
    Result := AnsiLowerCase(Str)
  else
    Result := LowerCase(Str);
end;

//______________________________________________________________________________
function GetVersion(var Line : string) : string;
var
  i : Integer;
  s : string;
  sl : TStringList;
  ext : string;
  old : string;
begin
  Result := '';

  sl := TStringList.Create;
  try

    try

      Split(sl, Line, '|');

      if sl.Count = 3 then Result := sl[0] + '.' + Strings[16];
      if sl.Count <> 6 then Exit;

      //Le résultat sera 'UNKNOWN' si le process ne va pas à son terme
      Result := sl[0] + '.' + Strings[14];

      if UseCache then
      begin
        if not CachePages.TryGetValue(sl[1], s) then
        begin
          //On récupère le code de la page
          s := GetPage(sl[1]);
          s := DoLowerCase(s);
          CachePages.Add(sl[1], s);
        end;
      end else begin
        //On récupère le code de la page
        s := GetPage(sl[1]);
        s := DoLowerCase(s);
      end;

      //On recherche le point de départ
      i := Pos(DoLowerCase(sl[2]), s);
      if i = 0 then Exit;
      Delete(s, 1, i + Length(sl[2]) - 1);

      //On recherche la borne de gauche
      i := Pos(DoLowerCase(sl[3]), s);
      if i = 0 then Exit;
      Delete(s, 1, i + Length(sl[3]) - 1);

      //On recherche la borne de droite
      i := Pos(DoLowerCase(sl[4]), s);
      if i = 0 then Exit;
      //On stocke dans s le n° de version trouvé
      Delete(s, i, Length(s));

      s := Replace(s, #13, '');
      s := Replace(s, #10, '');
      s := Trim(s);

      ext := Strings[12];

      //si le num a changé, on change l'extension en 'NEW' et on affiche le changement
      if (s <> '') and (sl[5] <> s) then begin
        ext := Strings[11];
        old := sl[5];
        sl[5] := s;
        s := old + ' >> ' + s;
        Line := merge(sl, '|');
      end;

      //si la ligne est correcte, on l'affiche dans TC
      if (s <> '') and (Length(s) < 100) and (Pos(#13, s) = 0) and (Pos(#10, s) = 0) then
        Result := sl[0] + ' (' + s + ').' + ext;

    except
      //En cas d'erreur pendant le process, on affiche une ligne 'ERREUR'
      Result := Strings[13];
    end;

  finally
    FreeAndNil(sl);
  end;

end;
//______________________________________________________________________________
function FsFindFirst(Path :PAnsiChar; var FindData : tWIN32FINDDATA) : THandle; stdcall;
begin
  Result := 0;
end;

function FsFindFirstW(path: PWideChar; var FindData:tWIN32FINDDATAW):thandle; stdcall;
var
  s : string;
begin
  FillChar(FindData, SizeOf(FindData), 0);
  CurrentPath := StrPas(Path);

  if UseCache then
    CachePages.Clear;

  CreateFileList;
  if FileList.Count = 0 then Result := INVALID_HANDLE_VALUE
  else begin
    Result := 0;
    FLIndex := 0;
    if Pos('.list', CurrentPath) > 0 then begin
      if ProgressBar(PluginNumber, PChar(Strings[3]), PChar(Copy(FileList[0], 1, Pos('|', FileList[0]) - 1)), 0) = 1 then
        Result := INVALID_HANDLE_VALUE;

      s := FileList[0];
      BuildFindData(FindData, GetVersion(s), Integer(FileList.Objects[0]));
      FileList[0] := s;
    end
    else BuildFindData(FindData, FileList[0], Integer(FileList.Objects[0]));
  end;

end;
//______________________________________________________________________________
function FsFindNext(Hdl : THandle; var FindData : TWIN32FINDDATA) : bool; stdcall;
begin
  Result := False;
end;

function FsFindNextW(Hdl:THandle; var FindDataW:tWIN32FINDDATAW):bool; stdcall;
var
  s : string;
  v : string;
begin
  FillChar(FindDataW, SizeOf(FindDataW), 0);
  Inc(FLIndex);
  Result := FileList.Count > FLIndex;
  if Result then begin
    if Pos('.list', CurrentPath) > 0 then begin
      if ProgressBar(PluginNumber, PChar(Strings[3]), PChar(Copy(FileList[FLindex], 1,
                     Pos('|', FileList[FLindex]) - 1)), Trunc(100 * (FLIndex / FileList.Count))) = 1 then
        Result := False;
      s := FileList[FLindex];
      v := GetVersion(s);
      FileList[FLindex] := s;
      BuildFindData(FindDataW, v, Integer(FileList.Objects[FLIndex]));
    end
    else BuildFindData(FindDataW, FileList[FLindex], Integer(FileList.Objects[FLIndex]));
  end
  else begin
   ProgressBar(PluginNumber, PChar(''), PChar(''), 100);
  end;
end;
//______________________________________________________________________________
function FsFindClose(Hdl : thandle) : Integer; stdcall;
var
  s : string;
begin
  Result := 0;
  if Pos('.list', CurrentPath) > 0 then begin
    s := CurrentPath;
    Delete(s, 1, 1);
    FileList.SaveToFile(PathExe + s);
  end;
end;
//______________________________________________________________________________
procedure FsGetDefRootName(DefRootName:PAnsiChar;MaxLen:integer); stdcall;
begin
  StrLCopy(DefRootName, 'Versions', MaxLen - 1);
end;
//______________________________________________________________________________
function FsGetFile(RemoteName, LocalName: PAnsiChar; CopyFlags: Integer;
  RemoteInfo: pRemoteInfo): Integer; stdcall;
begin
  Result := FS_FILE_OK;
end;

function FsGetFileW(RemoteName, LocalName: PWideChar; CopyFlags: integer;
  RemoteInfo: pRemoteInfo):integer; stdcall;
var
  Name : string;
begin
  Name := StrPas(RemoteName);

  if ExtractFileExt(Name) <> '.list' then begin
    Result := FS_FILE_NOTSUPPORTED;
  end
  else begin
    StrPCopy(RemoteName, PathExe + ExtractFileName(Name));
    CopyFile(PChar(PathExe + ExtractFileName(Name)), LocalName, False);
    Result := FS_FILE_OK;
  end;
end;
//______________________________________________________________________________
function FsPutFile(LocalName,RemoteName:pAnsichar;CopyFlags:integer):integer; stdcall;
begin
  Result := 0;
end;

function FsPutFileW(LocalName,RemoteName:pwidechar;CopyFlags:integer):integer; stdcall;
var
  Name : string;
begin
  Name := StrPas(RemoteName);

  if ExtractFileExt(Name) <> '.list' then begin
    Result := FS_FILE_NOTSUPPORTED;
  end
  else begin
    CopyFile(LocalName, PChar(PathExe + ExtractFileName(Name)), False);
    Result := FS_FILE_OK;
  end;
end;
//______________________________________________________________________________
function OnlyFileName(RemoteName : PChar) : string;
begin
  Result := StrPas(RemoteName);
  Result := DeleteUntilLast(Result, '\');
end;
//______________________________________________________________________________
function FsExtractCustomIcon(RemoteName:pAnsichar;ExtractFlags:integer;
  var TheIcon:hicon):integer; stdcall;
begin
  Result := FS_ICON_USEDEFAULT;
end;

function FsExtractCustomIconW(RemoteName:pwidechar;ExtractFlags:integer;
  var TheIcon:hicon):integer; stdcall;
var
  Name : string;
  Ofn : string;
  ext : string;
begin

  Name := StrPas(RemoteName);
  if Copy(Name, Length(Name) - 3, 4) = '\..\' then begin
    Result := FS_ICON_USEDEFAULT;
    Exit;
  end;

  Result := FS_ICON_EXTRACTED;
  Ofn := OnlyFileName(RemoteName);

  if Pos('\' + Strings[1], Name) = 1 then
    TheIcon := LoadIcon(HInstance, 'ZLANG')
  else if Pos('.list', Name) > 0 then begin
    ext := ExtractFileExt(Ofn);
    if      ext = '.' + Strings[11] then TheIcon := LoadIcon(HInstance, 'ZNEW')
    else if ext = '.' + Strings[16] then TheIcon := LoadIcon(HInstance, 'ZSRC')
    else if ext = '.' + Strings[12] then TheIcon := LoadIcon(HInstance, 'ZLINK')
    else if ext = '.' + Strings[14] then TheIcon := LoadIcon(HInstance, 'ZUNK')
    else if ext = '.list' then TheIcon := LoadIcon(HInstance, 'ZCONNECT')
    else if ofn = Strings[13] then TheIcon := LoadIcon(HInstance, 'ZERROR');
  end
  else if Name = '\' + Strings[10] then
    TheIcon := LoadIcon(HInstance, 'ZADD');
end;
//______________________________________________________________________________
function ExpandEnvVar(StringToExpand: string): string;
var
  Path: string;
begin
  Result := '';
  if StringToExpand = '' then
    Exit;
  SetLength(Path, 3200);
  ExpandEnvironmentStrings(PChar(StringToExpand), PChar(Path), Length(Path) - 1);
  SetLength(Path, StrLen(PChar(Path)));
  Result := ExpandFileName(Path);
end;
//______________________________________________________________________________
procedure CreateHTTPClient;
begin
  HTTPClient := THTTPClient.Create;
end;
//______________________________________________________________________________
procedure DLLEntryPoint(dwReason: DWORD);
begin
  if dwReason = DLL_PROCESS_ATTACH then
  begin
    FileList := TStringList.Create;
    PathExe := ExtractFilePath(GetDllPathName);
    ini := TIni.Create(PathExe + 'versions.ini');
    CreateHTTPClient;
    Strings := TLines.Create;
    Strings.SetLanguage(PathExe + Ini.GetS('Language') + '.lng');
    UseCache := Ini.GetB('UseCache', UseCache);
    if UseCache then
      CachePages := TDictionary<string, string>.Create;
    UseLowerCase := Ini.GetB('UseLowerCase', UseLowerCase);
  end
  else if dwReason = DLL_PROCESS_DETACH then
  begin
    try
      FreeAndNil(FileList);
      FreeAndNil(ini);
      FreeAndNil(Strings);
      if UseCache then
        FreeAndNil(CachePages);
      FreeAndNil(HTTPClient);
    except
    end;
  end;
end;

//______________________________________________________________________________

exports
  FsInit, FsInitW,
  FsFindFirst, FsFindFirstW,
  FsFindNext, FsFindNextW,
  FsFindClose, FsGetDefRootName,
  FsExecuteFile, FsExecuteFileW,
  FsExtractCustomIcon, FsExtractCustomIconW,
  FsGetFile, FsGetFileW,
  FsPutFile, FsPutFileW;

begin
  DllProc := @DLLEntryPoint;
  DisableThreadLibraryCalls(HInstance);
  DllEntryPoint(DLL_PROCESS_ATTACH);
end.
